/******************************************************************************
 * Fichier main :igmprt - An IGMP Proxy implementation
 ******************************************************************************
 * Fichier    : igmprt.c
 * Description: Implementation d'un proxy IGMP en se basant sur
 *              l'internet draft "draft-ietf-idmr-igmp-v3-07.txt", Mars 2001
 *              et "draft-ietf-idmr-igmp-proxy-01.txt", Janvier 2002
 * Date       : May 18, 2000
 * Auteurs    : wilbertdg@hetnet.nl
 *              lahmadi@loria.fr
 *              Anis.Ben-Hellel@loria.fr
 *              sachink@student.utdallas.edu
 * Last Modif : May 29, 2007
 *
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <syslog.h>
#include <pwd.h>
#include "conf.h"
#include "igmprt.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

/*version and isquerier variable from the config file*/
int version,querier, num_downstream;
char upstream_param[20];
char downstream_param[MAXVIFS][20];
int isblank(int c);
int process_cmd (char *cmd);

#define BUF_CMD	100

igmp_router_t router;
int go_on = 1;
char* pidfile = DEFAULT_PID_FILE_NAME;

static char str1[20], str2[20], printbuf[MAX_MSGBUFSIZE];

void igmp_info_print(igmp_router_t *router)
{
	char *pstr = printbuf;
	igmp_interface_t *ifp;
	igmp_group_t *gp;
	igmp_src_t *src;
	igmp_rep_t *rep;
	int nbytes; 

	nbytes = sprintf(pstr, "\nIGMP Table\n-----------------\n");
	pstr += nbytes;
	nbytes = sprintf(pstr, "\n%-14s %-9s %-14s %-7s %-14s %-14s\n",
		"interface","version","groups","mode","source","Members");
	pstr += nbytes;
	
	for (ifp=router->igmprt_interfaces; ifp; 
		ifp=(igmp_interface_t *)ifp->igmpi_next){
	  nbytes = sprintf(pstr, "%-14s 0x%x\n",inet_ntoa(ifp->igmpi_addr),
	  			ifp->igmpi_version);
	  pstr += nbytes;
	  if (ifp->igmpi_groups != NULL){
	    for(gp=ifp->igmpi_groups;gp;gp=(igmp_group_t*)gp->igmpg_next){
	      nbytes=sprintf(pstr,"%32s (%3d) %8s\n",inet_ntoa(gp->igmpg_addr),
      				gp->igmpg_timer,
	      	(gp->igmpg_fmode == IGMP_FMODE_INCLUDE)? "INCLUDE":"EXCLUDE");
	      pstr += nbytes;
	      if (gp->igmpg_sources != NULL) {
	        for (src=gp->igmpg_sources; src; 
				src=(igmp_src_t *)src->igmps_next) {
   		  nbytes = sprintf(pstr, "%54s (%d)\n",
		  	inet_ntoa(src->igmps_source), src->igmps_timer);
		  pstr += nbytes;
		}
	      }
	      if (gp->igmpg_members != NULL) {
		for (rep=gp->igmpg_members; rep; 
		 		rep=(igmp_rep_t *)rep->igmpr_next)
		  nbytes = sprintf(pstr, "%70s\n",inet_ntoa(rep->igmpr_addr));
		  pstr += nbytes;
		}
	      else {
	        nbytes = sprintf(pstr, "\n");
		pstr += nbytes;
	      }
	    }
	  } else {
	    nbytes = sprintf(pstr, "\n");
	    pstr += nbytes;
	  }
	}
}

/****************************************************************************
 *
 * LDAP communication
 *
 ***************************************************************************/
int validate(){

/* validation du report provenant d'un membre d'un groupe*/
  return 0;
}

/****************************************************************************
 *
 * Routines pour les enregistrements des membres: Ceci n'a pas ete specifie
 * dans le draft, mais pour se rendre compte vite q'un membre viend de quitter 
 * un groupe, ou pour garder trace du membres on peut utiliser ces routines
 *
 ***************************************************************************/
		    
/*
 * igmp_rep_t* igmp_rep_create()
 *
 * Cree un enregistrement d'un membre d'un groupe
 */
igmp_rep_t*
igmp_rep_create(struct in_addr srcaddr)
{
	igmp_rep_t* rep;

	if ((rep = (igmp_rep_t*) calloc(1, sizeof(*rep)))) {
	  rep->igmpr_addr.s_addr = srcaddr.s_addr;
	  rep->igmpr_next  = NULL;
	}
	return rep;
}

/*
 * void igmp_rep_cleanup()
 *
 * Effacer un enregistrement d'un membre 
 */
void
igmp_rep_cleanup(
	igmp_group_t *gp,
	igmp_rep_t * rep)
{
	igmp_rep_t *re;

	assert(rep != NULL);
	assert (gp != NULL);
	if (rep != gp->igmpg_members){
	  for (re=gp->igmpg_members;(igmp_rep_t *)re->igmpr_next != rep;
	  	re=(igmp_rep_t *)re->igmpr_next);		
	  re->igmpr_next = rep->igmpr_next;
	  free(rep);
	}else{
	  /*delete the head*/
	  re=gp->igmpg_members;
	  gp->igmpg_members = (igmp_rep_t *)re->igmpr_next;
	  free(re);
	}
	LOG((LOG_DEBUG, "igmp_rep_cleanup: %s\n", inet_ntoa(rep->igmpr_addr)));
}

/*
 * void igmp_rep_print()
 *
 * Imprimer un enregistrement d'un membre
 */
void
igmp_rep_print(
	igmp_rep_t* rep)
{
  printf("\t\tmembre:%-16s\n", 
	 inet_ntoa(rep->igmpr_addr));
}
/*
 * igmp_rep_t* igmp_group_rep_lookup()
 *
 * Lookup a membre in the sourcetable of a group
 */
igmp_rep_t*
igmp_group_rep_lookup(igmp_group_t *gp,
	 	      struct in_addr srcaddr)
{
  igmp_rep_t *rep;
  
  assert(gp != NULL);
  for (rep = gp->igmpg_members; rep; rep = rep->igmpr_next) {
    if (rep->igmpr_addr.s_addr == srcaddr.s_addr) {
      return rep;
    }
  }
  return NULL; 
}

/*
 * igmp_rep_t* igmp_group_rep_add()
 *
 * Add a member to the set of sources of a group
 */
igmp_rep_t*
igmp_group_rep_add(igmp_group_t *gp,
		   struct in_addr srcaddr)
{
	igmp_rep_t* rep;

	assert(gp != NULL);
	/* Return the source if it's already present */
	if ((rep = igmp_group_rep_lookup(gp, srcaddr))) {
	  return rep;
	}

	/* Create the source and add to the set */ 
	if ((rep = igmp_rep_create(srcaddr))) {
	  rep->igmpr_next = (igmp_rep_t *)gp->igmpg_members;
	  gp->igmpg_members = rep;
	}
	return rep;
}

/******************************************************************************
 *
 * igmp source routines 
 * 
 *****************************************************************************/
/*
 * igm_src_t* igmp_src_create()
 *
 * Cree un enregistrement d'une source
 */
igmp_src_t*
igmp_src_create(
	struct in_addr srcaddr)
{
	igmp_src_t* src;

	if ((src = (igmp_src_t*) calloc(1, sizeof(*src)))) {
	  src->igmps_source.s_addr = srcaddr.s_addr;
	  src->igmps_timer = 0;
	  src->igmps_fstate = TRUE;
	  src->igmps_next  = NULL;
	}
	return src;
}

/*
 * void igmp_src_cleanup()
 *
 * Effacer un enregistrement d'une source
 */
void
igmp_src_cleanup(
	igmp_group_t *gp,
	igmp_src_t * src)
{
	igmp_src_t *sr;

	assert(src != NULL);
	assert(gp != NULL);
	LOG((LOG_DEBUG, "igmp_src_cleanup: Source: %s\n", 
		inet_ntoa(src->igmps_source)));
	LOG((LOG_DEBUG, "igmp_src_cleanup: Group: %s\n", 
	 	inet_ntoa(gp->igmpg_addr)));
	if (src != gp->igmpg_sources){
	  for (sr=gp->igmpg_sources;(igmp_src_t *)sr->igmps_next != src;
	  	sr=(igmp_src_t *)sr->igmps_next);		
	  sr->igmps_next = src->igmps_next;
	  free(src);
	}else{
	  /*delete the head*/
	  sr=gp->igmpg_sources;
	  gp->igmpg_sources = (igmp_src_t *)sr->igmps_next;
	  free(sr);
	}
}

/*
 * void igmp_grp_print()
 *
 * Imprimer un enregistrement d'une source
 */
void
igmp_grp_print(char *group)
{
	struct in_addr gr;
	igmp_interface_t *ifp;
	igmp_src_t *src;
	igmp_rep_t *rep;
	igmp_group_t *gp;
	char *pstr = printbuf;
	int nbytes;

	gr.s_addr=inet_addr(group);
	for (ifp = (&router)->igmprt_interfaces; ifp; ifp = ifp->igmpi_next){
	  gp = igmprt_group_lookup(&router,ifp->igmpi_addr,gr);
	  nbytes = sprintf(pstr, "interface : %s\n", 
	  			inet_ntoa(ifp->igmpi_addr));
	  pstr += nbytes;
	  if (gp != NULL) { 
  	    nbytes = sprintf(pstr, "  %-16s %s %d\n", 
	    	inet_ntoa(gp->igmpg_addr),
		(gp->igmpg_fmode == IGMP_FMODE_EXCLUDE) ? "exclude" : "include",
		gp->igmpg_timer);
	    pstr += nbytes;
  	    if (gp->igmpg_sources != NULL) {
	      for (src=gp->igmpg_sources;src;src=src->igmps_next) {
	        nbytes = sprintf(pstr, "source : %s timer : %d\n",
				inet_ntoa(src->igmps_source), src->igmps_timer);
		pstr += nbytes;
	      }
	    }

   	    if (gp->igmpg_members != NULL) {
	      for (rep=gp->igmpg_members;rep;rep=rep->igmpr_next) {
	        nbytes = sprintf(pstr, "member : %s \n",
				inet_ntoa(rep->igmpr_addr));
		pstr += nbytes;
	      }
	    }
	  } else {
	    nbytes = sprintf(pstr, "unknown group\n");
	    pstr += nbytes;
	  }
	}
}
/*
 * void igmp_src_print()
 *
 * Imprimer un enregistrement d'une source
 */
void
igmp_src_print(char *source)
{
	struct in_addr srcaddr;
	igmp_interface_t *ifp;
	igmp_src_t *src;
	igmp_group_t *gp;
	char *pstr = printbuf;
	int nbytes;
	
	srcaddr.s_addr=inet_addr(source);

	for (ifp = (&router)->igmprt_interfaces; ifp; ifp = ifp->igmpi_next) {
	  for (gp = ifp->igmpi_groups; gp; gp = gp->igmpg_next) {
	    if ((src = igmp_group_src_lookup(gp, srcaddr)) != NULL) {
	      nbytes = sprintf(pstr, "interface : %s\n",
	    			inet_ntoa(ifp->igmpi_addr));
	      pstr += nbytes;
	      nbytes = sprintf(pstr, "\tgroup : %s\n",
	   			inet_ntoa(gp->igmpg_addr));
	      pstr += nbytes;

	      nbytes = sprintf(pstr, "\t\tsource:%-16s  %d\n", 
	 	inet_ntoa(src->igmps_source),src->igmps_timer);
	      pstr += nbytes;
	    }
	  }
	}
}
/*
 * igmp_src_t* igmp_group_src_lookup()
 *
 * Lookup a source in the sourcetable of a group
 */
igmp_src_t*
igmp_group_src_lookup(
	igmp_group_t *gp,
	struct in_addr srcaddr)
{
  igmp_src_t *src;
  
 if (gp != NULL){
  for (src = gp->igmpg_sources; src != NULL; src = (igmp_src_t *)src->igmps_next){
    if (src->igmps_source.s_addr == srcaddr.s_addr)
      return src;
  }
  }
  return NULL; 
}

/*
 * igmp_src_t* igmp_group_src_add()
 *
 * Add a source to the set of sources of a group
 */
igmp_src_t*
igmp_group_src_add(
	igmp_group_t *gp,
	struct in_addr srcaddr)
{
	igmp_src_t* src;

	assert(gp != NULL);
	/* Return the source if it's already present */
	if ((src = igmp_group_src_lookup(gp, srcaddr)))
	  return src;
	/* Create the source and add to the set */ 
	if ((src = igmp_src_create(srcaddr))) {
	  src->igmps_next = gp->igmpg_sources;
	  gp->igmpg_sources = src;
	}
	return src;
}



/******************************************************************************
 *
 * igmp group routines 
 *
 *****************************************************************************/

/*
 * igmp_group_t* igmp_group_create()
 *
 * Create a group record
 */
igmp_group_t*
igmp_group_create(
	struct in_addr groupaddr)
{
	igmp_group_t* gp;

	if ((gp = (igmp_group_t*) calloc(1, sizeof(*gp)))) {
		gp->igmpg_addr.s_addr = groupaddr.s_addr;
		gp->igmpg_fmode = IGMP_FMODE_INCLUDE;
		gp->igmpg_version = IGMP_VERSION_3;/*default version is V3*/
		gp->igmpg_timer = 0;
		gp->igmpg_sources = NULL;
		gp->igmpg_members = NULL;
		gp->igmpg_next = NULL;
		return gp;
	}else
		return NULL;
	
}

/*
 * void igmp_group_cleanup()
 *
 * Cleanup a group record
 */
void
igmp_group_cleanup(
	igmp_interface_t *ifp,
	igmp_group_t* gp)
{
	igmp_group_t *g;
  	igmp_rep_t *rep, *prep;
	assert(gp != NULL);
	assert(ifp != NULL);
	LOG((LOG_DEBUG, "igmp_group_cleanup: %s\n", inet_ntoa(gp->igmpg_addr)));

	/* Cleanup the members */
	for(prep = rep = gp->igmpg_members; prep; rep = prep) {
	  prep = prep->igmpr_next;
	  igmp_rep_cleanup(gp, rep);
	}

	if (ifp->igmpi_groups != gp){
	  g=ifp->igmpi_groups;
	  while((igmp_group_t *)g->igmpg_next != gp)		
	  g=(igmp_group_t *)g->igmpg_next;
	  g->igmpg_next=gp->igmpg_next;		
	  free(gp);
	}else{
	  /*delete the head*/
	  g=ifp->igmpi_groups;
	  ifp->igmpi_groups = (igmp_group_t *)g->igmpg_next;
	  free(g);
	} 
}

/*
 * void igmp_group_print()
 *
 * Print a group record
 */
void
igmp_group_print(igmp_group_t* gp)
{
  igmp_src_t *src;
  igmp_rep_t *rep;
  printf("  %-16s %s %d\n", 
	 inet_ntoa(gp->igmpg_addr),
	 (gp->igmpg_fmode == IGMP_FMODE_EXCLUDE) ? "exclude" : "include",
	 gp->igmpg_timer);
  if (gp->igmpg_sources != NULL)
    for (src=gp->igmpg_sources;src;src=(igmp_src_t *)src->igmps_next)
      printf("source : %s timer : %d\n",inet_ntoa(src->igmps_source),
      		src->igmps_timer);
   if (gp->igmpg_members != NULL)
    for (rep=gp->igmpg_members;rep;rep=(igmp_rep_t *)rep->igmpr_next)
      printf("member : %s \n",inet_ntoa(rep->igmpr_addr));
}

/******************************************************************************
 *
 * igmp interface routines 
 *
 *****************************************************************************/

/*
 * igmp_interface_t* igmp_interface_create()
 *
 * Create and initialize interface record
 */
igmp_interface_t*
igmp_interface_create(
	igmp_router_t *igmprt,
	struct in_addr ifaddr,
	char *ifname,
	vifi_t index)
{
  struct ifreq ifr;
  igmp_interface_t* ifp;
  struct ip_mreq mreq;
  char ra[4];
  int i;
  short flags;
  
	if (((ifp = (igmp_interface_t*) calloc(1, sizeof(*ifp)))) == NULL)
		return NULL;

	/* Allocate a buffer to receive igmp messages */ 
	ifp->igmpi_bufsize = get_interface_mtu(ifname);
	if (ifp->igmpi_bufsize == -1)
		ifp->igmpi_bufsize = MAX_MSGBUFSIZE;
	/* XXX Should make buffer slightly bigger for ip header */
	if ((ifp->igmpi_buf = (char*) calloc(1, ifp->igmpi_bufsize)) == NULL) {
		free(ifp);
		return NULL;
	}	

	/* Create a raw igmp socket */
	ifp->igmpi_socket = socket(AF_INET, SOCK_RAW, IPPROTO_IGMP);
	if (ifp->igmpi_socket == -1) {
		printf("can't create socket \n");
 		free(ifp->igmpi_buf);
		free(ifp);
		return NULL;
	}
	

	/* Initialize all fields */
	ifp->igmpi_addr.s_addr = ifaddr.s_addr;
	strncpy(ifp->igmpi_name, ifname, IFNAMSIZ);
	ifp->igmpi_groups = NULL;
	ifp->sch_group_query = NULL;

	if (strcmp(ifp->igmpi_name, upstream_param) == 0) {
	   ifp->igmpi_type = UPSTREAM;
	   upstream = ifp->igmpi_addr.s_addr;
	   igmprt->igmprt_upstream_vif = index;
	}else{
	   ifp->igmpi_type = DOWNSTREAM;
	}


	ifp->igmpi_isquerier = TRUE;
  	ifp->igmpi_version = version; /*IGMP_VERSION_3;*/
	ifp->igmpi_qi = IGMP_DEF_QI;
	ifp->igmpi_qri = IGMP_DEF_QRI;
	ifp->igmpi_rv = IGMP_DEF_RV;
	ifp->igmpi_gmi = ifp->igmpi_rv * ifp->igmpi_qi + ifp->igmpi_qri;
	ifp->igmpi_ti_qi = 0;
	ifp->igmpi_next = NULL;

	/* Set router alert */
	ra[0] = 148;
	ra[1] = 4;
	ra[2] = 0;
	ra[3] = 0;
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_OPTIONS, ra, 4);


	/* Set reuseaddr, ttl, loopback and set outgoing interface */
	i = 1;
	setsockopt(ifp->igmpi_socket, SOL_SOCKET, SO_REUSEADDR, 
		(void*)&i, sizeof(i));
	i = 1;
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_MULTICAST_TTL, 
		(void*)&i, sizeof(i));
	i = 1;
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_MULTICAST_LOOP, 
		(void*)&i, sizeof(i));
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_MULTICAST_IF, 
		(void*)&ifaddr, sizeof(ifaddr));
#ifdef Linux
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_PKTINFO, &i, sizeof(i));
#endif	
	//setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_RECVIF, &i, sizeof(i));
	/* Add membership to ALL_ROUTERS and ALL_ROUTERS_V3 on this interface */
	//if (ifp->igmpi_type == DOWNSTREAM) 
	{
	  mreq.imr_multiaddr.s_addr = inet_addr(IGMP_ALL_ROUTERS);
	  mreq.imr_interface.s_addr = ifaddr.s_addr;/*htonl(0);*/
	  setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, 
	  		(void*)&mreq, sizeof(mreq));
	  mreq.imr_multiaddr.s_addr = inet_addr(IGMP_ALL_ROUTERS_V3);
	  mreq.imr_interface.s_addr = ifaddr.s_addr;/*htonl(0);*/
	  setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, 
	  		(void*)&mreq, sizeof(mreq));
	}
	/* Tell the kernel this interface belongs to a multicast router */
	mrouter_onoff(ifp->igmpi_socket,1);
	ifp->igmpi_index = index;

	bzero(&ifr, sizeof(ifr));
	strcpy(ifr.ifr_name, ifname);
	if(ioctl(ifp->igmpi_socket, SIOCGIFINDEX, &ifr) < 0) {
		perror("Index");
	}
	ifp->igmpi_ifindex = ifr.ifr_ifru.ifru_ivalue;
	
	/* Set the interface flags to receive all multicast packets */
	ifp->igmpi_save_flags = get_interface_flags(ifname);
	if (ifp->igmpi_save_flags != -1) {
	  set_interface_flags(ifname, ifp->igmpi_save_flags | IFF_ALLMULTI);
	  /* If IFF_ALLMULTI didn't work, try IFF_PROMISC */
	  flags = get_interface_flags(ifname);
	  /* if (flags & IFF_ALLMULTI != IFF_ALLMULTI) */
	  set_interface_flags(ifname, ifp->igmpi_save_flags | IFF_PROMISC);
	}

	return ifp;
}

/*
 * void igmp_interface_cleanup()
 *
 * Cleanup an interface 
 */
void
igmp_interface_cleanup(igmp_interface_t* ifp)
{
	igmp_group_t* gp,*gp2;

	assert(ifp != NULL);

	/* Cleanup all groups */
	gp=ifp->igmpi_groups;
	ifp->igmpi_groups = NULL;
	while(gp != NULL){
		gp2=gp;
		gp=gp->igmpg_next;
		free(gp2);
	}
	/* Tell the kernel the multicast router is no more */
	mrouter_onoff(ifp->igmpi_socket, 0);
	close(ifp->igmpi_socket);

 	free(ifp->igmpi_buf);

	/* If we managed to get/set the interface flags, revert */
	if (ifp->igmpi_save_flags != -1)
		set_interface_flags(ifp->igmpi_name, ifp->igmpi_save_flags);
	free(ifp);
}

/*
 * igmp_group_t* igmp_interface_group_add()
 *
 * Add a group to the set of groups of an interface
 */
igmp_group_t*
igmp_interface_group_add(
	igmp_router_t* router,
	igmp_interface_t *ifp,
	struct in_addr groupaddr)
{
	igmp_group_t* gp;
	struct ip_mreq mreq;
	igmp_interface_t* upstream_interface;
	struct in_addr up;

	assert(ifp != NULL);

	/* Return the group if it's already present */
	if ((gp = igmp_interface_group_lookup(ifp, groupaddr)))
		return gp;
	/* Create the group and add to the set */ 
	if ((gp = igmp_group_create(groupaddr))) {
		gp->igmpg_next = ifp->igmpi_groups;
		ifp->igmpi_groups = gp;

		/*add group to the set of groups of upstream interface*/
		mreq.imr_multiaddr.s_addr=groupaddr.s_addr;
		mreq.imr_interface.s_addr=upstream;
		if ((VALID_ADDR(mreq.imr_multiaddr)) && 
		    (ifp->igmpi_addr.s_addr != upstream)) {
		  up.s_addr = upstream;
		  upstream_interface = igmprt_interface_lookup(router,up);
		  if (igmp_interface_group_lookup(upstream_interface,
		  	mreq.imr_multiaddr) == NULL) {
		    if (setsockopt(router->igmprt_up_socket, IPPROTO_IP, 
		        IP_ADD_MEMBERSHIP, (void *) &mreq, sizeof(mreq)) < 0) { 
		      setsockopt(router->igmprt_up_socket, IPPROTO_IP, 
	    		IP_DROP_MEMBERSHIP, (void *) &mreq, sizeof(mreq));
		      if (setsockopt(router->igmprt_up_socket, IPPROTO_IP, 
		      	IP_ADD_MEMBERSHIP, (void *) &mreq, sizeof(mreq)) < 0) { 
		        perror("setsockopt - IP_ADD_MEMBERSHIP");
		        exit(1);
		      }
		    }
		  }
		} 
	}
	return gp;
}

/*
 * igmp_group_t* igmp_interface_group_lookup()
 *
 * Lookup a group in the grouptable of an interface
 */
igmp_group_t*
igmp_interface_group_lookup(
	igmp_interface_t *ifp,
	struct in_addr groupaddr)
{
	igmp_group_t* gp;

	assert(ifp != NULL);
	for (gp = ifp->igmpi_groups; gp; gp = gp->igmpg_next)
		if (gp->igmpg_addr.s_addr == groupaddr.s_addr)
			return gp;
	return NULL; 
}

/*
 * void igmp_interface_membership_report_v12()
 *
 * Process a v1/v2 membership report
 */
void
igmp_interface_membership_report_v12(igmp_router_t* router,	
				     igmp_interface_t* ifp,
				     struct in_addr src,
				     igmpr_t* report,
				     int len)
{
	igmp_group_t* gp;
	igmp_rep_t *rep;

	/* Ignore a report for a non-multicast address */ 
	if (! IN_MULTICAST(ntohl(report->igmpr_group.s_addr))) {
	  return;
	}

	/* Find the group, and if not present, add it */
	if ((gp = igmp_interface_group_add(router,ifp, report->igmpr_group)) == 		NULL) {
	  return;
	}

    	/* find the member and add it if not present*/
	rep = igmp_group_rep_add(gp,src);

	/* Consider this to be a v3 is_ex{} report */
	igmp_group_handle_isex(router,ifp, gp, 0, NULL);
}
/*
 * void igmp_interface_membership_report_v3()
 *
 * Process a v3 membership report :only a group record in the report
 * len :will be used in the Baal implementation
 */
void
igmp_interface_membership_report_v3(igmp_router_t* router,
				    igmp_interface_t* ifp,
				    struct in_addr src,/* addresse member */
				    igmp_report_t* report,
				    int len)
{
	igmp_group_t* gp;
        igmp_rep_t *rep;
	u_short numsrc;
	u_short numgrps;
	u_char type;
	igmp_grouprec_t *grec = report->igmpr_group;
	int i;

	/* test l'interface: si report prevenant de l'upstream,
           et @addresse membre != mon addresse 	alors forward_upstream  = 1
         */
        if ((ifp->igmpi_addr.s_addr == upstream) && 
	    (src.s_addr != upstream) && 
	    (VALID_ADDR(report->igmpr_group[0].igmpg_group))) {
           forward_upstream = 1;
	}
	numgrps = ntohs(report->igmpr_numgrps);
	for (i=0; i < numgrps;i++){
		numsrc = ntohs(grec->igmpg_numsrc);
		/* Ignore a report for a non-multicast address */ 
		if (!IN_MULTICAST(ntohl(grec->igmpg_group.s_addr))) { 
		  LOG((LOG_DEBUG, "Not a multicast address %s\n", 
		  	inet_ntop(AF_INET, &grec->igmpg_group, str1, 20)));
		  grec =  (igmp_grouprec_t *) ((u_char *) grec + 
				sizeof(igmp_grouprec_t) +
				grec->igmpg_datalen * 4 +
				numsrc * sizeof(struct in_addr));
		  continue;
		}
		if ((ntohl(grec->igmpg_group.s_addr) & IGMP_LOCAL_GROUP) == 
		     IGMP_LOCAL_GROUP) {
		   if ((grec->igmpg_group.s_addr != IGMP_ALL_ROUTER) &&
		       (grec->igmpg_group.s_addr != IGMPV3_ALL_MCR) &&
		       (grec->igmpg_group.s_addr != IGMP_ALL_HOSTS)) {
		     grec =  (igmp_grouprec_t *) ((u_char *) grec + 
				sizeof(igmp_grouprec_t) +
				grec->igmpg_datalen * 4 +
				numsrc * sizeof(struct in_addr));
		     continue;
		   }
		}
		/* Find the group, and if not present, add it */
		if ((gp = igmp_interface_group_add(router,ifp, 
		  grec->igmpg_group)) == NULL) {
		  return;
		}
	        /* find the source of the report and add it if not present*/
		/*if (src.s_addr != ifp->igmpi_addr.s_addr) */
		rep=igmp_group_rep_add(gp,src);

		/* Find the group record type */
		type   = (u_char)grec->igmpg_type;
		switch(type) {
		case 1 : /* an IS_IN report */
		  LOG((LOG_DEBUG, "[%s] IS INCLUDE %d sources\n", 
		  	inet_ntop(AF_INET, &gp->igmpg_addr, str1, 20), numsrc));
		  igmp_group_handle_isin(router,ifp,gp,numsrc,
		  (struct in_addr *) &grec->igmpg_sources);
		  break;
		case 2 : /* an IS_EX report */
		  LOG((LOG_DEBUG, "[%s] IS EXCLUDE %d sources\n",
		  	inet_ntop(AF_INET, &gp->igmpg_addr, str1, 20), numsrc));
		  igmp_group_handle_isex(router,ifp,gp,numsrc,
		  (struct in_addr *) &grec->igmpg_sources);
		  break;
		case 3: /* an TO_IN report */
		  LOG((LOG_DEBUG, "[%s] TO INCLUDE %d sources\n",
		  	inet_ntop(AF_INET, &gp->igmpg_addr, str1, 20), numsrc));
		  igmp_group_handle_toin(router,ifp,gp,numsrc,
		  (struct in_addr *)&grec->igmpg_sources);
		  break;
		case 4: /* an TO_EX report */
		  LOG((LOG_DEBUG, "[%s] TO EXCLUDE %d sources\n",
		  	inet_ntop(AF_INET, &gp->igmpg_addr, str1, 20), numsrc));
		  igmp_group_handle_toex(router,ifp,gp,numsrc,
		  (struct in_addr *)&grec->igmpg_sources);
		  break;
		case 5: /* an ALLOW report */
		  LOG((LOG_DEBUG, "[%s] ALLOW NEW %d sources\n",
		  	inet_ntop(AF_INET, &gp->igmpg_addr, str1, 20), numsrc));
		  igmp_group_handle_allow(router,ifp,gp,numsrc,
		  (struct in_addr *)&grec->igmpg_sources);
		  break;
		case 6: /* an BLOCK report */
		  LOG((LOG_DEBUG, "[%s] BLOCK OLD %d sources\n",
		  	inet_ntop(AF_INET, &gp->igmpg_addr, str1, 20), numsrc));
		  igmp_group_handle_block(router,ifp,gp,numsrc,
		  (struct in_addr *)&grec->igmpg_sources);
		  break;
		default:
		  LOG((LOG_DEBUG, "igmp_interface_membership_report_v3: "
		  	"group record type undefined"));
		  break;
		}
		grec =  (igmp_grouprec_t *) ((u_char *) grec + 
				sizeof(igmp_grouprec_t) +
				grec->igmpg_datalen * 4 +
				numsrc * sizeof(struct in_addr));
	}

}


/*
 * void igmp_interface_print()
 *
 * Print status of an interface 
 */
void
igmp_interface_print(igmp_interface_t* ifp)
{
	igmp_group_t* gp;

	printf(" interface %s, %s ver=0x%x, index: %d ifindex: %d\n",
	 	inet_ntoa(ifp->igmpi_addr),
		(ifp->igmpi_type == UPSTREAM)?"UPSTREAM":"DOWNSTREAM", 
		ifp->igmpi_version, ifp->igmpi_index, ifp->igmpi_ifindex);
	for (gp = ifp->igmpi_groups; gp; gp = gp->igmpg_next)
		igmp_group_print(gp);
}

/******************************************************************************
 *
 * igmp router routines
 *
 *****************************************************************************/

/*
 * int igmprt_init()
 *
 * Initialize igmp router
 */
int
igmprt_init(
	igmp_router_t* igmprt)
{
	igmprt->igmprt_interfaces = NULL;
	igmprt->igmprt_thr_timer = igmprt->igmprt_thr_input = 0;
	igmprt->igmprt_flag_timer = 0;
	igmprt->igmprt_flag_input = 0;
	/*
	 * create datagram socket to tell igmpv3 host about reports to 
	 * send on upstream interface
	 */
	igmprt->igmprt_up_socket = socket( AF_INET, SOCK_DGRAM, 0 );
        if( igmprt->igmprt_up_socket < 0) {
	  perror("can't open upstream socket");
	  exit (1);
	}
	/*create raw socket to update mfc and vif table*/
	igmprt->igmprt_socket = socket(AF_INET, SOCK_RAW, IPPROTO_IGMP);
	if (igmprt->igmprt_socket < 0) 
	  perror("can't open igmp socket");
        forward_upstream = 0;
	return 1;
}

/*
 * void igmprt_destroy()
 *
 * Cleanup the router 
 */
void
igmprt_destroy(igmp_router_t* igmprt)
{
	igmp_interface_t *ifp, *ifp2;
	igmp_rep_t *rep, *prep;
	igmp_group_t *gp;
	membership_db *mem, *pmem;

	for (ifp2 = ifp = igmprt->igmprt_interfaces; ifp2; ifp = ifp2) {
	  /* Cleanup group member records */
	  for (gp = ifp->igmpi_groups; gp; gp = gp->igmpg_next) {
  	    for (prep = rep = gp->igmpg_members; prep; rep = prep) {
	      prep = prep->igmpr_next;
	      free(rep);
	    }
	  }
	  ifp2 = ifp2->igmpi_next;
	  igmp_interface_cleanup(ifp);
	}

	/* Cleanup membership database */
 	for (pmem = mem = igmprt->igmprt_membership_db; pmem; mem = pmem) {
	  pmem = pmem->next;
	  free(mem);
	}
        syslog(LOG_INFO, "Stopping igmp proxy");
	igmprt_stop(igmprt);
	
}

void
done(int sig)
{
	igmprt_destroy(&router);
	unlink(pidfile);
	exit(0);
}

/*
 * interface_t* igmprt_interface_lookup()
 *
 * Lookup a group, identified by the interface address
 */
igmp_interface_t*
igmprt_interface_lookup(
	igmp_router_t* igmprt, 
	struct in_addr ifaddr)
{
	igmp_interface_t *ifp;

	for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next)
		if (ifp->igmpi_addr.s_addr == ifaddr.s_addr)
			return ifp;
	return NULL;
}

/*
 * interface_t* igmprt_interface_lookup_index()
 *
 * Lookup a group, identified by the interface index
 */
igmp_interface_t*
igmprt_interface_lookup_index(
	igmp_router_t* igmprt, 
	int ifp_index)
{
	igmp_interface_t *ifp;

	for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next)
		if (ifp->igmpi_ifindex == ifp_index)
			return ifp;
	return NULL;
}

/*
 * igmp_entry_t* igmprt_group_lookup() 
 *
 * Lookup a group, identified by the interface and group address
 */
igmp_group_t*
igmprt_group_lookup(
	igmp_router_t* igmprt, 
	struct in_addr ifaddr, 
	struct in_addr groupaddr)
{
	igmp_interface_t *ifp; 

	if ((ifp = igmprt_interface_lookup(igmprt, ifaddr)))
		return igmp_interface_group_lookup(ifp, groupaddr);
	return NULL;
}

/*
 * struct igmp_interface_t* igmprt_interface_add()
 *
 * Add an interface to the interfacetable
 */
igmp_interface_t*
igmprt_interface_add(
	igmp_router_t* igmprt, 
	struct in_addr ifaddr,
	char *ifname,
	vifi_t index)
{
	igmp_interface_t *ifp;

	/* Return the interface if it's already in the set */
	if ((ifp = igmprt_interface_lookup(igmprt, ifaddr)))
		return ifp;
	/* Create the interface and add to the set */
	if ((ifp = igmp_interface_create(igmprt, ifaddr, ifname,index))) {
		ifp->igmpi_next = igmprt->igmprt_interfaces;
		igmprt->igmprt_interfaces = ifp;
	}
	return ifp;
}

/*
 * igmp_group_t* igmprt_group_add()
 *
 * Add a group to the grouptable of some interface
 */
igmp_group_t*
igmprt_group_add(
	igmp_router_t* igmprt, 
	struct in_addr ifaddr, 
	struct in_addr groupaddr)
{
	igmp_interface_t *ifp; 

	if ((ifp = igmprt_interface_lookup(igmprt, ifaddr)))
		return NULL;
	return igmp_interface_group_add(igmprt,ifp, groupaddr);
}  

/*
 * void igmprt_timergroup(igmp_router_t* igmprt)
 *
 * Decrement timers and handle whatever has to be done when one expires
 */
void
igmprt_timer(igmp_router_t* igmprt)
{
  igmp_interface_t* ifp;
  struct in_addr zero;
  
  zero.s_addr = 0;
  
  /* Handle every interface */
  for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next) {
    /* If we're the querier for this network, handle all querier 
     * duties */
    if (ifp->igmpi_isquerier == TRUE) {
      /* Deal with the general query */
      if (--ifp->igmpi_ti_qi <= 0) {
	ifp->igmpi_ti_qi = ifp->igmpi_qi;
	igmprt_membership_query(igmprt, ifp, &zero, NULL, 0, 0);
      }
    }else{
     /* If not the querier, deal with other-querier-present timer*/
     igmprt_timer_querier(ifp);	
    }
    /*handle group timer*/
    /*LOG((LOG_DEBUG, "handle timer group\n"));*/
    igmprt_timer_group(igmprt,ifp);
    /*handle source timer*/
    /*LOG((LOG_DEBUG, "handle timer source\n"));*/
    igmprt_timer_source(igmprt,ifp);
    /*handle scheduled query*/
    send_sh_query(igmprt,ifp);
   	
  }
}

/*
 * void* igmprt_timer_thread(void* arg)
 *
 * The thread that deals with the timer 
 */
void*
igmprt_timer_thread(void* arg)
{
  igmp_router_t* igmprt = (igmp_router_t*) arg;
  
  while (igmprt->igmprt_flag_timer) {
    igmprt_timer(igmprt);
    /* Should be changed to take care of drifting */
    usleep(1000000);
  }
  pthread_exit(NULL);
  return NULL;
}

void print_packet(char *buf, int len)
{
  	unsigned short *pi, i;
	printf("Received a packet: %d bytes\n0x%04x: ", len, 0);
        pi = (unsigned short *) buf;
        for (i=0; i < len/2; i++) {
            if (i > 0 && i%8 == 0) {
                printf("\n0x%04x: ", i*2);
            }
            printf("%04x ", htons(pi[i])); 
	} 
	printf("\n");
}

/*
 * void igmprt_input()
 *
 * Process an incoming igmp message
 */
void
igmprt_input(igmp_router_t* igmprt, igmp_interface_t* ifp)
{

	struct msghdr msg;
	struct iovec iov;
	struct cmsghdr *cmsg;
	char ctrl[1024];
	
	int if_index = -1;
	struct ip *iph;
	unsigned char *ptype;
	int n, igmplen;
	igmp_report_t *report;
	igmpv3q_t *query;
        struct in_addr src;
	int srsp;
	igmp_interface_t *ifpi;

	/* Read the igmp message */
	iov.iov_base = (char *)ifp->igmpi_buf;
	iov.iov_len = ifp->igmpi_bufsize;
	msg.msg_iov = &iov;
	msg.msg_iovlen= 1;
	msg.msg_control = ctrl;
	msg.msg_controllen = sizeof(ctrl);
       
	n=recvmsg(ifp->igmpi_socket,&msg,0);
	for(cmsg=CMSG_FIRSTHDR(&msg);cmsg!=NULL;cmsg =CMSG_NXTHDR(&msg,cmsg)) {
#ifdef Linux		
	  if (cmsg->cmsg_level == IPPROTO_IP && cmsg->cmsg_type == IP_PKTINFO){
	    struct in_pktinfo *i = (struct in_pktinfo *)CMSG_DATA(cmsg);
	    if_index = i->ipi_ifindex;
	    break;
	  }
#else		
	  if (cmsg->cmsg_type == IP_RECVIF){
	    if_index = CMSG_IFINDEX(cmsg);/* returns 1 or 2*/
	    if_index --; /* for us interfaces are 0 or 1*/
  	  }
#endif		
	}
	
	if(if_index != ifp->igmpi_ifindex) {
	  return;
	}

	if (n <= sizeof(*iph))
	  return;

	/* Set pointer to start of report */
	iph = (struct ip*) ifp->igmpi_buf;
	if ((igmplen = n - (iph->ip_hl << 2)) < IGMP_MINLEN)
	  return;

	src=iph->ip_src;
 	ptype = (unsigned char *) (ifp->igmpi_buf + (iph->ip_hl << 2));


	/*lookup the network interface from which the packet arrived*/
	/*if_index =  ifp->igmpi_index;*/
	ifpi = igmprt_interface_lookup_index(igmprt, if_index);
	if (ifpi == NULL) {
	  return;
	} 

	/* Handle the message */
	switch (*ptype) {
	case IGMP_MEMBERSHIP_QUERY:
	  LOG((LOG_DEBUG, "Received Membership query on %d %s\n", 
			if_index, inet_ntoa(ifpi->igmpi_addr)));
	  query = (igmpv3q_t *)(ifp->igmpi_buf + (iph->ip_hl << 2));
	  srsp=IGMP_SRSP(query);
	  if (query->igmpq_code == 0){
	    /*version 1 query*/
	    LOG((LOG_DEBUG, "igmpv1 query \n"));
	    receive_membership_query(ifpi,query->igmpq_group,NULL,
	      			   src.s_addr,0,srsp);
	  } else if (igmplen == 8){
	    /*version 2 query*/
	    LOG((LOG_DEBUG, "igmpv2 query \n"));
	    receive_membership_query(ifpi,query->igmpq_group,NULL,
		  			   src.s_addr,0,srsp);
	  } else if (igmplen >= 12){
	    /*version 3 query*/
	    receive_membership_query(ifpi,query->igmpq_group,
	    query->igmpq_sources,src.s_addr,query->igmpq_numsrc,srsp);
	  }
	  break;
	case IGMP_V1_MEMBERSHIP_REPORT:
	case IGMP_V2_MEMBERSHIP_REPORT:
	  //if (ifpi->igmpi_addr.s_addr != upstream)
	  {  
	    LOG((LOG_DEBUG, "Recieved Membership report v1 v2 on %d", 
			if_index));
	    LOG((LOG_DEBUG, ", %s\n", inet_ntoa(ifpi->igmpi_addr)));
	    igmp_interface_membership_report_v12(igmprt, ifpi, src, 
	  					    (igmpr_t*) ptype, igmplen);
	  }
	  break;
	case IGMP_V3_MEMBERSHIP_REPORT:
	  //if (ifpi->igmpi_addr.s_addr != upstream)
	  {
	        struct in_addr sin;
		memcpy(&sin, &src, sizeof(sin));
	  	LOG((LOG_DEBUG, "Recieved v3 report source %s on %d len %d",
	  		inet_ntop(AF_INET, &sin, str2, 20),if_index, igmplen));
	  	LOG((LOG_DEBUG, ", %s %s\n",
	  		inet_ntop(AF_INET, &ifpi->igmpi_addr, str2, 20),
	  		inet_ntop(AF_INET, &upstream, str1, 20)));
	  	report = (igmp_report_t *)(ifp->igmpi_buf + (iph->ip_hl << 2));
	  	igmp_interface_membership_report_v3(igmprt, ifpi, src, report,
							sizeof(report));
		LOG((LOG_DEBUG, "Done handling the report\n"));
	  }
	  break;
	default:
	  break;
	}
}

/*
 * void* igmprt_debug_thread(void* arg)
 *
 * Wait on all interfaces for packets to arrive
 */
char dbg_buf[20];
void*
igmprt_debug_thread(void* arg)
{
	int fd, nbytes, outfd;
	if (access(IGMP_PROXY_DEBUG_FIFO_OUT, F_OK) != 0) {
	  fd = mkfifo(IGMP_PROXY_DEBUG_FIFO_OUT, 0777);
	  if (fd < 0) {
            syslog(LOG_ERR, "unable to create debug fifo, code=%d (%s)",
                   errno, strerror(errno) );
	    done(0);
	  }
	}

	outfd = open(IGMP_PROXY_DEBUG_FIFO_OUT, O_RDWR);
	if (outfd < 0) {
          syslog(LOG_ERR, "unable to open debug fifo, code=%d (%s)",
                 errno, strerror(errno) );
	  done(0);
	}

	if (access(IGMP_PROXY_DEBUG_FIFO, F_OK) != 0) {
	  fd = mkfifo(IGMP_PROXY_DEBUG_FIFO, 0777);
	  if (fd < 0) {
            syslog(LOG_ERR, "unable to create debug fifo, code=%d (%s)",
                   errno, strerror(errno) );
	    done(0);
	  }
	}

	fd = open(IGMP_PROXY_DEBUG_FIFO, O_RDWR);
	if (fd < 0) {
          syslog(LOG_ERR, "unable to open debug fifo, code=%d (%s)",
                 errno, strerror(errno) );
	  done(0);
	}

	while(go_on) {
	  nbytes = read(fd, dbg_buf, 20);
	  if (nbytes < 0) {
	    continue;
	  }
	  go_on = process_cmd(dbg_buf);
	  nbytes = write(outfd, printbuf, MAX_MSGBUFSIZE);
	  if (nbytes <= 0) {
	    perror("write");
	  }
	}
	
	done(0);
	return NULL;
}

/*
 * void* igmprt_input_thread(void* arg)
 *
 * Wait on all interfaces for packets to arrive
 */
void*
igmprt_input_thread(void* arg)
{
	igmp_router_t* igmprt = (igmp_router_t*) arg;
	igmp_interface_t* ifp;
	fd_set allset, rset;
	int n, maxfd;

	/* Add the sockets from all interfaces to the set */
	FD_ZERO(&allset);
	maxfd = 0;
	for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next) {
		FD_SET(ifp->igmpi_socket, &allset);
		if (maxfd < ifp->igmpi_socket)
			maxfd = ifp->igmpi_socket;
	}

	if (maxfd == 0) {
	  LOG((LOG_INFO, "error: no interfaces available to wait for input\n"));
	  return NULL;
	}

	/* Wait for data on all sockets */
	while (1) {
	  FD_COPY(&allset, &rset);
	  n = select(maxfd+1, &rset, NULL, NULL, NULL);
	  if (n == 0)
	    sleep(1);
	  for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next) {
	    if (FD_ISSET(ifp->igmpi_socket, &rset)) {
	      igmprt_input(igmprt, ifp);
	      if (--n == 0)
		break;
	    }
	  } 
	}
	return NULL;
}

/*
 * void igmprt_start(igmp_router_t* igmprt)
 *
 * Start the threads of this router 
 */
void
igmprt_start(igmp_router_t* igmprt)
{
	int err;

	/* Return if already running */
	if (igmprt->igmprt_running)
		return;

	/* Create and start the timer handling (thread) */
	igmprt->igmprt_flag_timer = 1;
	igmprt->igmprt_thr_timer = 0;
	if ((err = pthread_create(&igmprt->igmprt_thr_timer, NULL, 
		igmprt_timer_thread, (void*) igmprt)) != 0)
		printf("couldn't start timer thread\n");

	/* Create and start input handling (thread) */
	igmprt->igmprt_flag_input = 1;
	igmprt->igmprt_thr_input = 0;
	if ((err = pthread_create(&igmprt->igmprt_thr_input, NULL, 
		igmprt_input_thread, (void*) igmprt)) != 0)
		printf("couldn't start input thread\n");

	igmprt->igmprt_running = 1;
}

/*
 * void igmprt_stop(igmp_router_t* igmprt)
 *
 * Stop the threads of this router 
 */
void
igmprt_stop(igmp_router_t* igmprt)
{
	/* Return if not running */ 
	if (! igmprt->igmprt_running)
		return;

	k_stop_proxy(igmprt->igmprt_socket);

	/* Signal threads to stop */
	igmprt->igmprt_flag_timer = 0;
	igmprt->igmprt_flag_input = 0;

	/* Wait for the threads to finish */
	/* XXX Should use attach */ 
	igmprt->igmprt_running = 0;
	/* Make sure select of  input-thread wakes up */	
	/*if ((ifp = igmprt->igmprt_interfaces))
		write(ifp->igmpi_socket, &i, sizeof(i));*/
	sleep(3);
}

/*
 * void igmprt_print()
 *
 * Print the status of the igmpv3 proxy/router
 */
void
igmprt_print(igmp_router_t* igmprt)
{
	igmp_interface_t* ifp;

	assert(igmprt != NULL);
	printf("igmp router:\n");
	for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next)
		igmp_interface_print(ifp);
}

/******************************************************************************
 *
 * The application
 *
 *****************************************************************************/

int
write_pid()
{
	FILE *fp = fopen(pidfile, "r");
	if (fp) {
	  char pid_string[11];

	  /* PID lock file exists */
	  if (fgets(pid_string, sizeof(pid_string), fp) != NULL) {
	    int oldpid;

	    oldpid = (int) strtol(pid_string, (char **)NULL, 10);
	    if (oldpid) {
		if (kill(oldpid, 0) == 0) {
		    /* gateway process already running */
		    return oldpid;
		}
	    }
	  }
	  fclose(fp);
	  unlink(pidfile);
	}

	/*
	 * create new pid lock file 
	 */
	fp = fopen(pidfile, "w");
    	if (fp == NULL) {
	  return -1;
	}
    	
	fprintf(fp, "%d\n", getpid());
	fclose(fp);
	return 0;
}


void parse_option(void)
{
  
  FILE *f;
  char linebuf[100];
  char *w, *s;
  int option;
  
  if ((f=fopen(DEFAULT_CONF_FILE_PATH,"r")) == NULL)
    printf("can't open config file DEFAULT_CONF_FILE_PATH\n");
  
  while(fgets(linebuf,sizeof(linebuf),f) != NULL){
    s=linebuf;
    w=(char *)next_word(&s);
    option=wordToOption(w);
    switch(option){
    case EMPTY:
      continue;
      break;
    case IGMPVERSION:
      syslog(LOG_INFO, "igmp version:");
      if (EQUAL((w=(char *)next_word(&s))," ")){
	version= DEFAULT_VERSION;
	syslog(LOG_INFO, "version set to default :0x%x\n",DEFAULT_VERSION);
      }
      else if(sscanf(w,"%d",&version) == 1){
	syslog(LOG_INFO, "0x%x\n",version);
      }
      break;
    case IS_QUERIER:
      syslog(LOG_INFO, "isquerier: ");
      if (EQUAL((w=(char *)next_word(&s))," ")){
	querier=DEFAULT_ISQUERIER;
	syslog(LOG_INFO, "is_querier set to default :%d\n",DEFAULT_ISQUERIER);
      }
      else if(sscanf(w,"%d",&querier) == 1)
	syslog(LOG_INFO, "%d\n",querier);
	break;
    case UPSTREAM:
        w=(char *)next_word(&s);
        strcpy(upstream_param, w);
	break;
    case DOWNSTREAM:
    {
	int i = 0;
        while(*s && i < MAXVIFS)
	{
	  w = (char *)next_word(&s);
	  strcpy(downstream_param[i], w);
	  i++;
	}
	num_downstream = i;
	break;
    }
    default:
      break;
    }
  }
}

void print_usage()
{
	char *pstr = printbuf;
	int nbytes;
	nbytes = sprintf(pstr, "g <multicast adress>:show a group state\n");
	pstr += nbytes;
	nbytes = sprintf(pstr, "s <unicast address> :show a source state\n");
	pstr += nbytes;
	nbytes = sprintf(pstr, "a		    :show the IGMP table\n");
	pstr += nbytes;
	nbytes = sprintf(pstr, "q		    :stop igmp proxy\n");
}

/* Process a command */
int process_cmd (char *cmd)
{
	char *line, *pstr = printbuf;
	char group[20], source[20];
	int nbytes; 

	memset(printbuf, 0, 20);

	/* skip white spaces */
	line = cmd + 1;
	while (isblank(*line)) line ++;
	switch (*cmd) {
	  case '?':
	    print_usage();
	    break;
	  case 'q':
	      nbytes = sprintf(pstr, "Exiting\n");
	    return 0;
	  case 'g':
	    /* print a group details*/
	    if (sscanf(line,"%s",group) != 1){
	      nbytes = sprintf(pstr, "enter valid multicast address\n");
	    } else {
	      igmp_grp_print(group);
	    }
	    break;
	  case 's':
	    /* print a source of a group details*/
	    if (sscanf(line,"%s",source) !=1){	
	      nbytes = sprintf(pstr, "enter valid unicast address\n");
	    } else {
	      igmp_src_print(source);
	    }
	    break;
	  case 'a' :
	    /*print all details of the router*/
	    igmp_info_print(&router); 
	    break;
	  default:
	    print_usage();
	    break;
        }
	return 1;
}		

void
usage(char *name) 
{
    fprintf(stderr, "usage: %s [-d] [-v]\n", name);
    fprintf(stderr, "\t-d : Run IGMP proxy as a daemon\n");
    fprintf(stderr, "\t-v : Version of the IGMP proxy\n");
    exit(1);
}

static void daemonize( const char *lockfile )
{
    pid_t pid;

    /* Fork off the parent process */
    pid = fork();
    if (pid < 0) {
        syslog( LOG_ERR, "unable to fork daemon, code=%d (%s)",
                errno, strerror(errno) );
        exit(0);
    } else if (pid > 0) {
        /* If we got a good PID, then we can exit the parent process. */
        exit(0);
    }

    /* Change the current working directory.  This prevents the current
       directory from being locked; hence not being able to remove it. */
    if ((chdir("/")) < 0) {
        syslog( LOG_ERR, "unable to change directory to %s, code %d (%s)",
                "/", errno, strerror(errno) );
        exit(0);
    }
}

int
main(int argc, char *argv[])
{
	int run_as_daemon = 0;
	struct sockaddr_in *psin;
	interface_list_t *ifl, *ifp;
	char cmd[BUF_CMD], lockfile[BUF_CMD];
	vifi_t vifi;
	int ch, rc;

	/* Initialize the logging interface */
	openlog(argv[0], LOG_PID, LOG_LOCAL5 );
	syslog(LOG_INFO, "Starting" );
	
	bzero(str1, 20);
	bzero(str2, 20);
	while ((ch = getopt(argc, argv, "hdv?")) != EOF) {
	  switch (ch) {
	  case 'v':
	  case 'V':
	    printf("IGMP Proxy version: %s\n", IGMP_PROXY_VERSION);
	    return 0;
	  case 'd':
	  case 'D':
	    run_as_daemon = 1;
	    break;
	  default:
	    usage(argv[0]);
	    break;
	  }
	}

	/*parse option on config file*/
	parse_option();

	if (run_as_daemon) {
	  sprintf(lockfile, "/var/lock/subsys/%s", argv[0]);
	  daemonize(lockfile);
	}

	/* Initialize */
	signal(SIGUSR1, done);
	signal(SIGKILL, done);
	signal(SIGABRT, done);
	signal(SIGTERM, done);
	signal(SIGHUP, done);

	rc = write_pid();
	if (rc < 0) {
	  fprintf(stderr, "%s: Can't create pid lockfile.\n", argv[0]);
	  exit(1);
        } else if (rc) {
	  /* already running */
	  fprintf(stderr, "%s: already running, pid %d\n", argv[0], rc);
	  exit(1);
        } 

	if ((rc = pthread_create(&router.igmprt_thr_debug, NULL, 
		igmprt_debug_thread, (void*) &router)) != 0)
	  printf("couldn't start debug thread\n");

	/* Create and initialize the router */
	igmprt_init(&router);
	k_init_proxy(((igmp_router_t *) &router)->igmprt_socket);

	numvifs = 0;
	/* Add all the multicast enabled ipv4 interfaces */
	ifl = get_interface_list(AF_INET, IFF_MULTICAST, IFF_LOOPBACK);
	for (vifi=0,ifp=ifl;ifp;ifp=ifp->ifl_next,vifi++) {
	  psin = (struct sockaddr_in*) &ifp->ifl_addr;
	  igmprt_interface_add(&router, psin->sin_addr, ifp->ifl_name,vifi);
	  k_proxy_add_vif(((igmp_router_t *) &router)->igmprt_socket,
	  		  psin->sin_addr.s_addr,vifi);
  	  numvifs++;
	}
	free_interface_list(ifl);
	
	/* Print the status of the router */
	//igmprt_print(&router);
	
	/* Start the router */
	igmprt_start(&router);

 	if (!run_as_daemon) {
	  while (go_on) {
	    /* Read and process commands */
	    printf("> ");
	    fflush(stdout);
	    if (fgets(cmd, BUF_CMD, stdin) != NULL) {
	      go_on = process_cmd(cmd);
	      printf("%s", printbuf);
	    }
	  }
	  /* Done */
	  done(0);
	} else {
	  while(1);
	}	
	
	return 0;
}
